import asyncio

import numpy as np
import rerun as rr
from nova import Nova
from scipy.spatial.transform import Rotation as R

from nova_rerun_bridge import colors
from nova_rerun_bridge.dh_robot import DHRobot
from nova_rerun_bridge.hull_visualizer import HullVisualizer
from nova_rerun_bridge.robot_visualizer import RobotVisualizer

# Constants
RECORDING_INTERVAL = 0.016  # 16ms per point
TIME_INTERVAL_NAME = f"time_interval_{RECORDING_INTERVAL}"


def log_safety_zones_once(motion_group: str, optimizer_config, robot: DHRobot):
    """
    Log hull outlines for the safety zones defined in the optimizer configuration.
    """
    mounting_transform = optimizer_config.mounting

    for zone in optimizer_config.safety_setup.safety_zones:
        geom = zone.geometry
        zone_id = zone.id
        entity_path = f"{motion_group}/zones/zone_{zone_id}"

        if geom.compound is not None:
            child_geoms = geom.compound.child_geometries
            polygons = HullVisualizer.compute_hull_outlines_from_geometries(child_geoms)
        elif geom.convex_hull is not None:

            class ChildWrapper:
                pass

            c = ChildWrapper()
            c.convex_hull = geom.convex_hull
            polygons = HullVisualizer.compute_hull_outlines_from_geometries([c])
        else:
            polygons = []

        accumulated = robot.pose_to_matrix(mounting_transform)
        polygons = apply_transform_to_polygons(polygons, accumulated)

        # Log polygons as wireframe outlines
        if polygons:
            line_segments = [p.tolist() for p in polygons]  # convert numpy arrays to lists
            rr.log(
                entity_path,
                rr.LineStrips3D(
                    line_segments, radii=rr.Radius.ui_points(0.75), colors=[[221, 193, 193, 255]]
                ),
                static=True,
                timeless=True,
            )


def apply_transform_to_polygons(polygons, transform):
    """
    Apply a transformation matrix to a list of polygons.
    """
    transformed_polygons = []
    for polygon in polygons:
        # Convert polygon to homogeneous coordinates
        homogeneous_polygon = np.hstack((polygon, np.ones((polygon.shape[0], 1))))
        # Apply the transformation
        transformed_polygon = np.dot(transform, homogeneous_polygon.T).T
        # Convert back to 3D coordinates
        transformed_polygons.append(transformed_polygon[:, :3])
    return transformed_polygons


def log_joint_positions_once(motion_group: str, robot: DHRobot, joint_position):
    """Compute and log joint positions for a robot."""
    joint_positions = robot.calculate_joint_positions(joint_position)
    line_segments = [
        [joint_positions[i], joint_positions[i + 1]] for i in range(len(joint_positions) - 1)
    ]
    segment_colors = [colors.colors[i % len(colors.colors)] for i in range(len(line_segments))]

    rr.log(
        f"{motion_group}/dh_parameters",
        rr.LineStrips3D(line_segments, colors=segment_colors),
        static=True,
    )


class MotionGroupProcessor:
    def __init__(self):
        self.last_tcp_pose = {}

    def tcp_pose_changed(self, motion_group: str, tcp_pose) -> bool:
        """Check if the TCP pose has changed compared to the last logged value."""
        last_pose = self.last_tcp_pose.get(motion_group)
        current_position = [tcp_pose.position.x, tcp_pose.position.y, tcp_pose.position.z]
        current_rotation = [tcp_pose.orientation.x, tcp_pose.orientation.y, tcp_pose.orientation.z]

        if last_pose is None:
            # No previous pose, consider it as changed
            self.last_tcp_pose[motion_group] = (current_position, current_rotation)
            return True

        last_position, last_rotation = last_pose
        if current_position != last_position or current_rotation != last_rotation:
            # Update the cache and return True if either position or rotation has changed
            self.last_tcp_pose[motion_group] = (current_position, current_rotation)
            return True

        return False

    def log_tcp_orientation(self, motion_group: str, tcp_pose):
        """Log TCP orientation and position."""
        rotation_vector = [tcp_pose.orientation.x, tcp_pose.orientation.y, tcp_pose.orientation.z]
        rotation = R.from_rotvec(rotation_vector)
        angle = rotation.magnitude() if rotation.magnitude() != 0 else 0.0
        axis_angle = rotation.as_rotvec() / angle if angle != 0 else [0, 0, 0]

        rr.log(
            f"{motion_group}/tcp_position",
            rr.Transform3D(
                translation=[tcp_pose.position.x, tcp_pose.position.y, tcp_pose.position.z],
                rotation=rr.RotationAxisAngle(axis=axis_angle, angle=angle),
            ),
            static=True,
        )


async def process_motion_group_state():
    """Process motion group states and log data efficiently."""
    print("Starting motion group state processing...", flush=True)

    nova = Nova()
    motion_group_infos_api = nova._api_client.motion_group_infos_api
    motion_group_api = nova._api_client.motion_group_api
    processor = MotionGroupProcessor()

    rr.init(application_id="nova", recording_id="nova_live", spawn=True)
    rr.set_time_seconds(TIME_INTERVAL_NAME, 0)

    # Fetch active motion groups
    motion_groups = await motion_group_api.list_motion_groups("cell")

    print(f"Active motion groups: {motion_groups.instances}", flush=True)

    for active_motion_group in motion_groups.instances:
        motion_group = active_motion_group.motion_group
        optimizer_config = await motion_group_infos_api.get_optimizer_configuration(
            "cell", motion_group
        )

        # Initialize DHRobot and Visualizer
        robot = DHRobot(optimizer_config.dh_parameters, optimizer_config.mounting)
        visualizer = RobotVisualizer(
            robot=robot,
            robot_model_geometries=optimizer_config.safety_setup.robot_model_geometries,
            tcp_geometries=optimizer_config.safety_setup.tcp_geometries,
            static_transform=True,
            base_entity_path=motion_group,
            albedo_factor=[0, 255, 100],
            model_from_controller=active_motion_group.model_from_controller,
        )

        # Log safety zones (only once)
        log_safety_zones_once(motion_group, optimizer_config, robot)

        print(f"Subscribing to motion group {motion_group}", flush=True)
        async for state in motion_group_infos_api.stream_motion_group_state("cell", motion_group):
            if processor.tcp_pose_changed(motion_group, state.state.tcp_pose):
                # Log joint positions
                log_joint_positions_once(motion_group, robot, state.state.joint_position)

                # Log robot geometries
                visualizer.log_robot_geometry(state.state.joint_position)

                processor.log_tcp_orientation(motion_group, state.state.tcp_pose)

            await asyncio.sleep(0.01)  # Prevents CPU overuse

    await nova._api_client.close()


if __name__ == "__main__":
    asyncio.run(process_motion_group_state())



from enum import Enum, auto
from typing import Dict, List

import numpy as np
import rerun as rr
from nova.api import models
from scipy.spatial.transform import Rotation

from nova_rerun_bridge.collision_scene import extract_link_chain_and_tcp
from nova_rerun_bridge.consts import TIME_INTERVAL_NAME
from nova_rerun_bridge.dh_robot import DHRobot
from nova_rerun_bridge.robot_visualizer import RobotVisualizer


class TimingMode(Enum):
    """Controls how trajectories are timed relative to each other."""

    RESET = auto()  # Start at time_offset
    CONTINUE = auto()  # Start after last trajectory
    SYNC = auto()  # Use exact time_offset, don't update last time
    OVERRIDE = auto()  # Use exact time_offset and reset last time


# Track both last end time and last offset separately
_last_end_time = 0.0
_last_offset = 0.0


def log_motion(
    motion_id: str,
    model_from_controller: str,
    motion_group: str,
    optimizer_config: models.OptimizerSetup,
    trajectory: List[models.TrajectorySample],
    collision_scenes: Dict[str, models.CollisionScene],
    time_offset: float = 0,
    timing_mode: TimingMode = TimingMode.CONTINUE,
):
    """
    Fetch and process a single motion with timing control.

    Args:
        ...existing args...
        timing_mode: Controls how trajectory timing is handled
            RESET: Start at time_offset (default)
            CONTINUE: Start after last trajectory
            SYNC: Use exact time_offset provided
    """
    global _last_end_time, _last_offset

    # Calculate start time based on timing mode
    if timing_mode == TimingMode.CONTINUE:
        effective_offset = _last_end_time + _last_offset
    elif timing_mode == TimingMode.SYNC:
        effective_offset = _last_end_time
    elif timing_mode == TimingMode.OVERRIDE:
        effective_offset = time_offset
        _last_end_time = time_offset
    else:  # TimingMode.RESET
        effective_offset = time_offset
        _last_end_time = time_offset

    # Initialize DHRobot and Visualizer
    robot = DHRobot(optimizer_config.dh_parameters, optimizer_config.mounting)

    collision_link_chain, collision_tcp = extract_link_chain_and_tcp(collision_scenes)

    visualizer = RobotVisualizer(
        robot=robot,
        robot_model_geometries=optimizer_config.safety_setup.robot_model_geometries,
        tcp_geometries=optimizer_config.safety_setup.tcp_geometries,
        static_transform=False,
        base_entity_path=f"motion/{motion_group}",
        model_from_controller=model_from_controller,
        collision_link_chain=collision_link_chain,
        collision_tcp=collision_tcp,
    )

    rr.set_time_seconds(TIME_INTERVAL_NAME, effective_offset)

    # Process trajectory points
    log_trajectory(
        motion_id=motion_id,
        motion_group=motion_group,
        robot=robot,
        visualizer=visualizer,
        trajectory=trajectory,
        optimizer_config=optimizer_config,
        timer_offset=effective_offset,
    )

    # Update last times based on timing mode
    if trajectory:
        if timing_mode == TimingMode.SYNC:
            _last_offset = trajectory[-1].time
        else:
            _last_offset = 0
            _last_end_time = effective_offset + trajectory[-1].time


def log_trajectory_path(
    motion_id: str, trajectory: List[models.TrajectorySample], motion_group: str
):
    points = [
        [p.tcp_pose.position.x, p.tcp_pose.position.y, p.tcp_pose.position.z] for p in trajectory
    ]
    rr.log(
        f"motion/{motion_group}/trajectory",
        rr.LineStrips3D([points], colors=[[1.0, 1.0, 1.0, 1.0]]),
    )

    rr.log("logs/motion", rr.TextLog(f"{motion_group}/{motion_id}", level=rr.TextLogLevel.INFO))


def get_times_column(
    trajectory: List[models.TrajectorySample], timer_offset: float = 0
) -> rr.TimeSecondsColumn:
    times = np.array([timer_offset + point.time for point in trajectory])
    times_column = rr.TimeSecondsColumn(TIME_INTERVAL_NAME, times)
    return times_column


def log_trajectory(
    motion_id: str,
    motion_group: str,
    robot: DHRobot,
    visualizer: RobotVisualizer,
    trajectory: List[models.TrajectorySample],
    optimizer_config: models.OptimizerSetup,
    timer_offset: float,
):
    """
    Process a single trajectory point and log relevant data.
    """
    rr.set_time_seconds(TIME_INTERVAL_NAME, timer_offset)

    times_column = get_times_column(trajectory, timer_offset)

    log_trajectory_path(motion_id, trajectory, motion_group)

    # Calculate and log joint positions
    line_segments_batch = []
    for point in trajectory:
        joint_positions = robot.calculate_joint_positions(point.joint_position)
        line_segments_batch.append(joint_positions)

    rr.send_columns(
        f"motion/{motion_group}/dh_parameters",
        times=[times_column],
        components=[
            rr.LineStrips3D.indicator(),
            rr.components.LineStrip3DBatch(line_segments_batch),
            rr.components.ColorBatch([0.5, 0.5, 0.5, 1.0] * len(line_segments_batch)),
        ],
    )

    # Log the robot geometries
    visualizer.log_robot_geometries(trajectory, times_column)

    # Log TCP pose/orientation
    log_tcp_pose(trajectory, motion_group, times_column)

    # Log joint data
    log_joint_data(trajectory, motion_group, times_column, optimizer_config)

    # Log scalar data
    log_scalar_values(trajectory, motion_group, times_column, optimizer_config)


def log_tcp_pose(trajectory: List[models.TrajectorySample], motion_group, times_column):
    """
    Log TCP pose (position + orientation) data.
    """
    tcp_positions = []
    tcp_rotations = []

    # Collect data from the trajectory
    for point in trajectory:
        # Collect TCP position
        tcp_positions.append(
            [point.tcp_pose.position.x, point.tcp_pose.position.y, point.tcp_pose.position.z]
        )

        # Convert and collect TCP orientation as axis-angle
        rotation_vector = [
            point.tcp_pose.orientation.x,
            point.tcp_pose.orientation.y,
            point.tcp_pose.orientation.z,
        ]
        rotation = Rotation.from_rotvec(rotation_vector)
        angle = rotation.magnitude()
        axis_angle = rotation.as_rotvec() / angle if angle != 0 else [0, 0, 0]
        tcp_rotations.append(rr.RotationAxisAngle(axis=axis_angle, angle=angle))

    rr.send_columns(
        f"motion/{motion_group}/tcp_position",
        times=[times_column],
        components=[
            rr.Transform3D.indicator(),
            rr.components.Translation3DBatch(tcp_positions),
            rr.components.RotationAxisAngleBatch(tcp_rotations),
        ],
    )


def log_joint_data(
    trajectory: List[models.TrajectorySample],
    motion_group,
    times_column,
    optimizer_config: models.OptimizerSetup,
) -> None:
    """
    Log joint-related data (position, velocity, acceleration, torques) from a trajectory as columns.
    """
    # Initialize lists for each joint and each data type (assuming 6 joints)
    num_joints = 6
    joint_data = {
        "velocity": [[] for _ in range(num_joints)],
        "acceleration": [[] for _ in range(num_joints)],
        "position": [[] for _ in range(num_joints)],
        "torque": [[] for _ in range(num_joints)],
        "velocity_lower_limit": [[] for _ in range(num_joints)],
        "velocity_upper_limit": [[] for _ in range(num_joints)],
        "acceleration_lower_limit": [[] for _ in range(num_joints)],
        "acceleration_upper_limit": [[] for _ in range(num_joints)],
        "position_lower_limit": [[] for _ in range(num_joints)],
        "position_upper_limit": [[] for _ in range(num_joints)],
        "torque_limit": [[] for _ in range(num_joints)],
    }

    # Collect data from the trajectory
    for point in trajectory:
        for i in range(num_joints):
            joint_data["velocity"][i].append(point.joint_velocity.joints[i])
            joint_data["acceleration"][i].append(point.joint_acceleration.joints[i])
            joint_data["position"][i].append(point.joint_position.joints[i])
            if point.joint_torques and len(point.joint_torques.joints) > i:
                joint_data["torque"][i].append(point.joint_torques.joints[i])

            # Collect joint limits
            joint_data["velocity_lower_limit"][i].append(
                -optimizer_config.safety_setup.global_limits.joint_velocity_limits[i]
            )
            joint_data["velocity_upper_limit"][i].append(
                optimizer_config.safety_setup.global_limits.joint_velocity_limits[i]
            )
            joint_data["acceleration_lower_limit"][i].append(
                -optimizer_config.safety_setup.global_limits.joint_acceleration_limits[i]
            )
            joint_data["acceleration_upper_limit"][i].append(
                optimizer_config.safety_setup.global_limits.joint_acceleration_limits[i]
            )
            joint_data["position_lower_limit"][i].append(
                optimizer_config.safety_setup.global_limits.joint_position_limits[i].lower_limit
            )
            joint_data["position_upper_limit"][i].append(
                optimizer_config.safety_setup.global_limits.joint_position_limits[i].upper_limit
            )
            if point.joint_torques and len(point.joint_torques.joints) > i:
                joint_data["torque_limit"][i].append(
                    optimizer_config.safety_setup.global_limits.joint_torque_limits[i]
                )

    # Send columns if data is not empty
    for data_type, data in joint_data.items():
        for i in range(num_joints):
            if data[i]:
                rr.send_columns(
                    f"motion/{motion_group}/joint_{data_type}_{i + 1}",
                    times=[times_column],
                    components=[rr.components.ScalarBatch(data[i])],
                )


def log_scalar_values(
    trajectory: List[models.TrajectorySample],
    motion_group,
    times_column,
    optimizer_config: models.OptimizerSetup,
):
    """
    Log scalar values such as TCP velocity, acceleration, orientation velocity/acceleration, time, and location.
    """
    scalar_data = {
        "tcp_velocity": [],
        "tcp_acceleration": [],
        "tcp_orientation_velocity": [],
        "tcp_orientation_acceleration": [],
        "time": [],
        "location_on_trajectory": [],
        "tcp_velocity_limit": [],
        "tcp_orientation_velocity_lower_limit": [],
        "tcp_orientation_velocity_upper_limit": [],
        "tcp_acceleration_lower_limit": [],
        "tcp_acceleration_upper_limit": [],
        "tcp_orientation_acceleration_lower_limit": [],
        "tcp_orientation_acceleration_upper_limit": [],
    }

    # Collect data from the trajectory
    for point in trajectory:
        if point.tcp_velocity is not None:
            scalar_data["tcp_velocity"].append(point.tcp_velocity)
        if point.tcp_acceleration is not None:
            scalar_data["tcp_acceleration"].append(point.tcp_acceleration)
        if point.tcp_orientation_velocity is not None:
            scalar_data["tcp_orientation_velocity"].append(point.tcp_orientation_velocity)
        if point.tcp_orientation_acceleration is not None:
            scalar_data["tcp_orientation_acceleration"].append(point.tcp_orientation_acceleration)
        if point.time is not None:
            scalar_data["time"].append(point.time)
        if point.location_on_trajectory is not None:
            scalar_data["location_on_trajectory"].append(point.location_on_trajectory)
        if optimizer_config.safety_setup.global_limits.tcp_velocity_limit is not None:
            scalar_data["tcp_velocity_limit"].append(
                optimizer_config.safety_setup.global_limits.tcp_velocity_limit
            )
        if optimizer_config.safety_setup.global_limits.tcp_orientation_velocity_limit is not None:
            scalar_data["tcp_orientation_velocity_lower_limit"].append(
                -optimizer_config.safety_setup.global_limits.tcp_orientation_velocity_limit
            )
            scalar_data["tcp_orientation_velocity_upper_limit"].append(
                optimizer_config.safety_setup.global_limits.tcp_orientation_velocity_limit
            )
        if optimizer_config.safety_setup.global_limits.tcp_acceleration_limit is not None:
            scalar_data["tcp_acceleration_lower_limit"].append(
                -optimizer_config.safety_setup.global_limits.tcp_acceleration_limit
            )
            scalar_data["tcp_acceleration_upper_limit"].append(
                optimizer_config.safety_setup.global_limits.tcp_acceleration_limit
            )
        if (
            optimizer_config.safety_setup.global_limits.tcp_orientation_acceleration_limit
            is not None
        ):
            scalar_data["tcp_orientation_acceleration_lower_limit"].append(
                -optimizer_config.safety_setup.global_limits.tcp_orientation_acceleration_limit
            )
            scalar_data["tcp_orientation_acceleration_upper_limit"].append(
                optimizer_config.safety_setup.global_limits.tcp_orientation_acceleration_limit
            )

    # Send columns if data is not empty
    for key, values in scalar_data.items():
        if values:
            rr.send_columns(
                f"motion/{motion_group}/{key}",
                times=[times_column],
                components=[rr.components.ScalarBatch(values)],
            )


def to_trajectory_samples(self) -> List[models.TrajectorySample]:
    """Convert JointTrajectory to list of TrajectorySample objects."""
    samples = []
    for joint_pos, time, location in zip(self.joint_positions, self.times, self.locations):
        sample = models.TrajectorySample(
            joint_position=joint_pos, time=time, location_on_trajectory=location
        )
        samples.append(sample)
    return samples



import asyncio

from nova import MotionSettings
from nova.actions import jnt, ptp
from nova.core.nova import Nova
from nova.types import Pose

from nova_rerun_bridge import NovaRerunBridge


async def test():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()

        cell = nova.cell()
        controllers = await cell.controllers()
        controller = controllers[0]

        # Connect to the controller and activate motion groups
        async with controller[0] as motion_group:
            home_joints = await motion_group.joints()
            tcp_names = await motion_group.tcp_names()
            tcp = tcp_names[0]

            # Get current TCP pose and offset it slightly along the x-axis
            current_pose = await motion_group.tcp_pose(tcp)
            target_pose = current_pose @ Pose((1, 0, 0, 0, 0, 0))

            actions = [
                jnt(home_joints),
                ptp(target_pose),
                jnt(home_joints),
                ptp(target_pose @ [100, 0, 0, 0, 0, 0]),
                jnt(home_joints),
                ptp(target_pose @ (100, 100, 0, 0, 0, 0)),
                jnt(home_joints),
                ptp(target_pose @ Pose((0, 100, 0, 0, 0, 0))),
                jnt(home_joints),
            ]

            # you can update the settings of the action
            for action in actions:
                action.settings = MotionSettings(tcp_velocity_limit=200)

            joint_trajectory = await motion_group.plan(actions, tcp)

            await bridge.log_trajectory(joint_trajectory, tcp, motion_group)
            await motion_group.execute(joint_trajectory, tcp, actions=actions)

        await nova.close()


if __name__ == "__main__":
    asyncio.run(test())



import subprocess
import tempfile
from pathlib import Path
from zipfile import ZipFile

import requests


def get_project_root() -> Path:
    """Get the root directory of the user's project"""
    # Start from current working directory and look for common project markers
    cwd = Path.cwd()
    markers = ["pyproject.toml", "setup.py", ".git"]

    current = cwd
    while current != current.parent:
        if any((current / marker).exists() for marker in markers):
            return current
        current = current.parent

    # Fallback to cwd if no project root found
    return cwd


def get_current_version(models_dir: Path) -> str:
    version_file = models_dir / "version.txt"
    if version_file.exists():
        return version_file.read_text().strip()
    return ""


def get_latest_release_version() -> str:
    api_url = (
        "https://api.github.com/repos/wandelbotsgmbh/wandelbots-js-react-components/releases/latest"
    )
    response = requests.get(api_url)
    if response.status_code != 200:
        raise Exception(f"Failed to get latest release: {response.status_code}")
    return response.json()["tag_name"].lstrip("v")


def download_and_extract(version: str, tmp_dir: Path) -> Path:
    zip_url = f"https://github.com/wandelbotsgmbh/wandelbots-js-react-components/archive/refs/tags/v{version}.zip"
    print(f"Downloading {zip_url}...")

    response = requests.get(zip_url)
    if response.status_code != 200:
        raise Exception(f"Failed to download: {response.status_code}")

    zip_path = tmp_dir / f"{version}.zip"
    zip_path.write_bytes(response.content)

    with ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(tmp_dir)

    return tmp_dir / f"wandelbots-js-react-components-{version}/public/models"


def check_gltf_transform():
    try:
        subprocess.run(["gltf-transform", "--version"], check=True, capture_output=True)
    except (subprocess.CalledProcessError, FileNotFoundError):
        raise RuntimeError(
            "gltf-transform not found. Install with: npm install -g @gltf-transform/cli"
        )


def decompress_glb(input_path: Path, output_path: Path):
    """Process GLB using gltf-transform"""
    temp_path = output_path.with_suffix(".temp.glb")

    # First decompress draco
    subprocess.run(
        ["gltf-transform", "dedup", str(input_path), str(temp_path)],
        check=True,
        capture_output=True,
    )

    # Then convert to unlit
    subprocess.run(
        ["gltf-transform", "unlit", str(temp_path), str(output_path)],
        check=True,
        capture_output=True,
    )

    temp_path.unlink()


def update_robot_models():
    check_gltf_transform()
    version = get_latest_release_version()
    models_dir = Path.cwd() / "models"
    current_version = get_current_version(models_dir)

    if version == current_version:
        print("Models are up to date")
        return

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        new_models_dir = download_and_extract(version, tmp_path)

        # Create models directory if it doesn't exist
        models_dir.mkdir(parents=True, exist_ok=True)

        # Process all GLB files
        for glb_file in new_models_dir.rglob("*.glb"):
            output_path = models_dir / glb_file.name
            print(f"Processing {glb_file.name}...")
            decompress_glb(glb_file, output_path)

        # Write version file
        (models_dir / "version.txt").write_text(version)

        print(f"Successfully updated robot models {current_version} ➜ {version}")


if __name__ == "__main__":
    update_robot_models()



from typing import Any, Dict, List, Tuple

import numpy as np
import rerun as rr
import trimesh
from nova.api import models
from scipy.spatial.transform import Rotation

from nova_rerun_bridge import colors
from nova_rerun_bridge.conversion_helpers import normalize_pose
from nova_rerun_bridge.hull_visualizer import HullVisualizer


def log_collision_scenes(collision_scenes: Dict[str, models.CollisionScene]):
    for scene_id, scene in collision_scenes.items():
        entity_path = f"collision_scenes/{scene_id}"
        for collider_id, collider in scene.colliders.items():
            log_colliders_once(entity_path, {collider_id: collider})


def log_colliders_once(entity_path: str, colliders: Dict[str, models.Collider]):
    for collider_id, collider in colliders.items():
        pose = normalize_pose(collider.pose)

        if collider.shape.actual_instance.shape_type == "sphere":
            rr.log(
                f"{entity_path}/{collider_id}",
                rr.Ellipsoids3D(
                    radii=[
                        collider.shape.actual_instance.radius,
                        collider.shape.actual_instance.radius,
                        collider.shape.actual_instance.radius,
                    ],
                    centers=[[pose.position.x, pose.position.y, pose.position.z]],
                    colors=[(221, 193, 193, 255)],
                ),
                timeless=True,
            )

        elif collider.shape.actual_instance.shape_type == "box":
            rr.log(
                f"{entity_path}/{collider_id}",
                rr.Boxes3D(
                    centers=[[pose.position.x, pose.position.y, pose.position.z]],
                    half_sizes=[
                        collider.shape.actual_instance.size_x,
                        collider.shape.actual_instance.size_y,
                        collider.shape.actual_instance.size_z,
                    ],
                    colors=[(221, 193, 193, 255)],
                ),
                timeless=True,
            )

        elif collider.shape.actual_instance.shape_type == "capsule":
            height = collider.shape.actual_instance.cylinder_height
            radius = collider.shape.actual_instance.radius

            # Generate trimesh capsule
            capsule = trimesh.creation.capsule(height=height, radius=radius, count=[6, 8])

            # Extract vertices and faces for solid visualization
            vertices = np.array(capsule.vertices)

            # Transform vertices to world position
            transform = np.eye(4)
            transform[:3, 3] = [pose.position.x, pose.position.y, pose.position.z - height / 2]
            rot_mat = Rotation.from_rotvec(
                np.array([pose.orientation.x, pose.orientation.y, pose.orientation.z])
            )
            transform[:3, :3] = rot_mat.as_matrix()

            vertices = np.array([transform @ np.append(v, 1) for v in vertices])[:, :3]

            polygons = HullVisualizer.compute_hull_outlines_from_points(vertices)

            if polygons:
                line_segments = [p.tolist() for p in polygons]
                rr.log(
                    f"{entity_path}/{collider_id}",
                    rr.LineStrips3D(
                        line_segments,
                        radii=rr.Radius.ui_points(0.75),
                        colors=[[221, 193, 193, 255]],
                    ),
                    static=True,
                    timeless=True,
                )

        elif collider.shape.actual_instance.shape_type == "convex_hull":
            polygons = HullVisualizer.compute_hull_outlines_from_points(
                collider.shape.actual_instance.vertices
            )

            if polygons:
                line_segments = [p.tolist() for p in polygons]
                rr.log(
                    f"{entity_path}/{collider_id}",
                    rr.LineStrips3D(
                        line_segments, radii=rr.Radius.ui_points(1.5), colors=[colors.colors[2]]
                    ),
                    static=True,
                    timeless=True,
                )

                vertices, triangles, normals = HullVisualizer.compute_hull_mesh(polygons)

                rr.log(
                    f"{entity_path}/{collider_id}",
                    rr.Mesh3D(
                        vertex_positions=vertices,
                        triangle_indices=triangles,
                        vertex_normals=normals,
                        albedo_factor=[colors.colors[0]],
                    ),
                    static=True,
                    timeless=True,
                )


def extract_link_chain_and_tcp(collision_scenes: dict) -> Tuple[List[Any], List[Any]]:
    """Extract link chain and TCP from collision scenes."""
    # Get first scene (name can vary)
    scene = next(iter(collision_scenes.values()), None)
    if not scene:
        return [], []

    # Try to get motion groups
    motion_group = next(iter(scene.motion_groups.values()), None)
    if not motion_group:
        return [], []

    return (getattr(motion_group, "link_chain", []), getattr(motion_group, "tool", []))



import asyncio

from nova import Controller, MotionSettings, Nova
from nova.actions import jnt, ptp
from nova.types import Pose

from nova_rerun_bridge import NovaRerunBridge
from nova_rerun_bridge.trajectory import TimingMode

"""
Example: Move multiple robots simultaneously.

Prerequisites:
- A cell with two robots: one named "ur" and another named "kuka".
"""


async def move_robot(controller: Controller, bridge: NovaRerunBridge):
    async with controller[0] as motion_group:
        home_joints = await motion_group.joints()
        tcp_names = await motion_group.tcp_names()
        tcp = tcp_names[0]

        current_pose = await motion_group.tcp_pose(tcp)
        target_pose = current_pose @ (100, 0, 0, 0, 0, 0)
        actions = [jnt(home_joints), ptp(target_pose), jnt(home_joints)]

        trajectory = await motion_group.plan(actions, tcp)
        await bridge.log_trajectory(trajectory, tcp, motion_group, timing_mode=TimingMode.SYNC)

        await motion_group.plan_and_execute(actions, tcp)


async def main():
    nova = Nova()
    bridge = NovaRerunBridge(nova)
    await bridge.setup_blueprint()

    cell = nova.cell()
    controllers = await cell.controllers()
    controller = controllers[0]

    # Connect to the controller and activate motion groups
    async with controller[0] as motion_group:
        home_joints = await motion_group.joints()
        tcp_names = await motion_group.tcp_names()
        tcp = tcp_names[0]

        # Get current TCP pose and offset it slightly along the x-axis
        current_pose = await motion_group.tcp_pose(tcp)
        target_pose = current_pose @ Pose((1, 0, 0, 0, 0, 0))

        actions = [
            jnt(home_joints),
            ptp(target_pose),
            jnt(home_joints),
            ptp(target_pose @ [100, 0, 0, 0, 0, 0]),
            jnt(home_joints),
            ptp(target_pose @ (100, 100, 0, 0, 0, 0)),
            jnt(home_joints),
            ptp(target_pose @ Pose((0, 100, 0, 0, 0, 0))),
            jnt(home_joints),
        ]

        # you can update the settings of the action
        for action in actions:
            action.settings = MotionSettings(tcp_velocity_limit=200)

        joint_trajectory = await motion_group.plan(actions, tcp)

        await bridge.log_trajectory(joint_trajectory, tcp, motion_group)
        await bridge.log_trajectory(joint_trajectory, tcp, motion_group)

        await motion_group.execute(joint_trajectory, tcp, actions=actions)

    cell = nova.cell()
    ur = await cell.controller("ur")
    kuka = await cell.controller("kuka")
    await asyncio.gather(move_robot(ur, bridge=bridge), move_robot(kuka, bridge=bridge))

    # Connect to the controller and activate motion groups
    async with controller[0] as motion_group:
        home_joints = await motion_group.joints()
        tcp_names = await motion_group.tcp_names()
        tcp = tcp_names[0]

        # Get current TCP pose and offset it slightly along the x-axis
        current_pose = await motion_group.tcp_pose(tcp)
        target_pose = current_pose @ Pose((1, 0, 0, 0, 0, 0))

        actions = [
            jnt(home_joints),
            ptp(target_pose),
            jnt(home_joints),
            ptp(target_pose @ [100, 0, 0, 0, 0, 0]),
            jnt(home_joints),
            ptp(target_pose @ (100, 100, 0, 0, 0, 0)),
            jnt(home_joints),
            ptp(target_pose @ Pose((0, 100, 0, 0, 0, 0))),
            jnt(home_joints),
        ]

        # you can update the settings of the action
        for action in actions:
            action.settings = MotionSettings(tcp_velocity_limit=200)

        joint_trajectory = await motion_group.plan(actions, tcp)

        await bridge.log_trajectory(joint_trajectory, tcp, motion_group)
        await bridge.log_trajectory(joint_trajectory, tcp, motion_group)

        await motion_group.execute(joint_trajectory, tcp, actions=actions)

    await nova.close()
    await bridge.cleanup()


if __name__ == "__main__":
    asyncio.run(main())
